/*
 * A Driver For USB CardMan
 *
 * cardman_usb.c v1.0 Sat Nov 18 17:05:12 MET 2000 
 *
 * Fri Aug 17 19:57:18 MEST 2001 herp - cm6020 firmware bug workaround
 * Fri Aug 24 10:12:04 MEST 2001 herp - wrong module count fixed
 * Sat Jan 19 00:31:37 MEST 2002 herp - support for modversions
 * Sat Mar 16 02:05:01 MEST 2002 herp - fixed timeout, copy_from/to_user,
 *                                      and signedness (u_char vs. short)
 * Fri May 24 17:51:07 MEST 2002 herp - several bugfixes
 * Wed Jul 17 15:00:00 MEST 2002 mh   - added ioctl CARDOFF and changed powering behaviour
 * Mon Oct  7 11:09:59 GMT 2002  mh   - added PPS functionality
 * Mon Dec  9 18:13:46 GMT 2002  mh   - reflect changes in USB subssystem API for kernel version 2.4.20
 * Tue Jun 29 12:23:05 CEST 2004 ct/mh - reflect changes in USB subsystem for kernel 2.6.X, add T1MODE2 card handling
 *
 * current version: 2.4.0beta1
 *
 * (C) 2000,2001,2002,2003,2004 Omnikey AG
 * All rights reserved
 *
 */

#include <linux/config.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 #include <linux/errno.h>
 #include <linux/smp_lock.h>
 #include <linux/completion.h>
 #include <asm/uaccess.h>
#else
 #ifdef	CONFIG_MODVERSIONS
  #define MODVERSIONS
  #include <linux/modversions.h>
 #endif
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/ioctl.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/usb.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/delay.h>

#define	__CM2020__
#include	"../cardman.h"
  
#undef min
#define	min(a,b)	(a<b?a:b)

#undef	CM2020_DEBUG
#undef	DEBUG_2
#undef	ATR_DEBUG
#define	CM2020_DEBUG	0
//#define	DEBUG_2
//#define	ATR_DEBUG

#ifdef CM2020_DEBUG
static int cm_debug = CM2020_DEBUG;
MODULE_PARM(cm_debug, "i");
#define DEBUG(n, args...) if (cm_debug >= (n)) printk(KERN_INFO args)
#else
#define DEBUG(n, args...)
#endif //CM2020_DEBUG

#define	STATIC static

STATIC volatile char *version=
"cardman_usb.c v2.4.0beta1 support.linux@omnikey.com";

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define FILL_CONTROL_URB usb_fill_control_urb
#define FILL_INT_URB     usb_fill_int_urb
/* prevent races between open() and disconnect() */
static DECLARE_MUTEX (disconnect_sem);
#endif

static unsigned char ucFIDITable_T1_FW1FW2[10][14] =
{
//FI      00    01    02    03    04    05    06    07    08    09    10    11    12    13
//DI
/* 0 */{ 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11},
/* 1 */{ 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x91, 0x11, 0x11, 0x11, 0x11},
/* 2 */{ 0x12, 0x12, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x92, 0x11, 0x11, 0x11, 0x11},
/* 3 */{ 0x13, 0x13, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x93, 0x11, 0x11, 0x11, 0x11},
/* 4 */{ 0x14, 0x14, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x94, 0x11, 0x11, 0x11, 0x11},
/* 5 */{ 0x18, 0x18, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x98, 0x11, 0x11, 0x11, 0x11},
/* 6 */{ 0x18, 0x18, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x98, 0x11, 0x11, 0x11, 0x11},
/* 7 */{ 0x18, 0x18, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x98, 0x11, 0x11, 0x11, 0x11},
/* 8 */{ 0x18, 0x18, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x98, 0x11, 0x11, 0x11, 0x11},
/* 9 */{ 0x18, 0x18, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x98, 0x11, 0x11, 0x11, 0x11}
};

#define	LOCK_DEVICE	           0
#define	IS_ANY_T0	           1
#define	IS_ANY_T1	           2
#define T1MODE2                    3
#define DO_AUTOPPS                 6
#define	IS_OP_PENDING	           7
#define	IS_OPEN		           8
#define	IS_INVERSE	           9
#define	IS_ATR_PRESENT	           10
#define	IS_ATR_VALID	           11
#define	IS_BAD_CSUM	           14
#define	IS_BAD_CARD	           15
#define CANCEL_ALL                 16

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC int cmu_probe(struct usb_interface *,
			const struct usb_device_id *);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
STATIC void *cmu_probe(struct usb_device *,unsigned int,
			const struct usb_device_id *id);
#else
STATIC void *cmu_probe(struct usb_device *,unsigned int);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC void cmu_disconnect(struct usb_interface *);
#else
STATIC void cmu_disconnect(struct usb_device *,void *);
#endif

STATIC ssize_t cmu_read(struct file *,char *,size_t,loff_t *);
STATIC ssize_t cmu_write(struct file *,const char *,size_t,loff_t *);
STATIC int cmu_ioctl(struct inode *,struct file *,unsigned int,
				unsigned long);
STATIC int cmu_open(struct inode *,struct file *);
STATIC int cmu_close(struct inode *,struct file *);

STATIC struct file_operations cmu_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        .owner=		THIS_MODULE,
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        .owner=		NULL,
#endif
	.read=	        cmu_read,
	.write=		cmu_write,
	.ioctl=		cmu_ioctl,
	.open=		cmu_open,
	.release=	cmu_close,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#else
 STATIC usb_cardman_t *cardman_table[CM2020_MAX_DEV] = { NULL, };
#endif

// use support for "new-style" USB hotplugging (see usb.h for more information)
STATIC const struct usb_device_id cardman_id_table []  = {
  {USB_DEVICE (0x076b, 0x0596), }, // Cardman 2020
  {USB_DEVICE (0x076b, 0x1784), }, // Cardman 6020
  {USB_DEVICE (0x08D4, 0x0009), }, // FSC USB SmartCard Reader
   { }
};
MODULE_DEVICE_TABLE (usb, cardman_id_table);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static struct usb_class_driver cmu_class = {
	.name=	"usb/cm%d",
	.fops=	&cmu_fops,
	.mode=	S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH,
	.minor_base=	CM2020_MINOR,
};
#endif

STATIC struct usb_driver cmu_driver = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
  .owner=		THIS_MODULE,
#endif
	.name=		"cardman",
	.probe=		cmu_probe,
	.disconnect=	cmu_disconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#else
	.driver_list={NULL, NULL},
	.fops= &cmu_fops,
	.minor= CM2020_MINOR,
#endif
	.id_table=	cardman_id_table
};

enum {
	b_1111, b_1110, b_1101, b_1100,
	b_1011, b_1010, b_1001, b_1000,
	b_0111, b_0110, b_0101, b_0100,
	b_0011, b_0010, b_0001, b_0000
};

STATIC unsigned char irtab[16]={
	b_0000, b_1000, b_0100, b_1100,
	b_0010, b_1010, b_0110, b_1110,
	b_0001, b_1001, b_0101, b_1101,
	b_0011, b_1011, b_0111, b_1111
};

STATIC inline unsigned char invert_revert(unsigned char ch) {
	return (irtab[ch&0x0f]<<4) | irtab[ch>>4];
}

STATIC void str_invert_revert(unsigned char *b,int len) {
int i;
	for(i=0;i<len;i++)
		b[i]=invert_revert(b[i]);
}

STATIC inline int xusb_unlink_urb(struct urb *);

STATIC inline int xusb_submit_urb(struct urb *urb) {
int i;
	DEBUG(3, "->usb_submit_urb(urb=%p)\n",urb);

	urb->dev=((usb_cardman_t *)urb->context)->dev; /* !&!"/()!" */
        
        #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	  if (!in_interrupt() && (urb->dev != NULL) && 
             (((usb_cardman_t*)(urb->context))->irq->status == -EPIPE) ){
	    DEBUG(4, "Clearing halt on IRQ pipe and resubmit irq urb in submit urb\n");
	    usb_clear_halt(urb->dev, ((usb_cardman_t*)(urb->context))->irq->pipe);
	    usb_submit_urb(((usb_cardman_t*)(urb->context))->irq, GFP_KERNEL);
	  }
        #endif
	
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	  if (!in_interrupt()){
	    i=usb_submit_urb(urb, GFP_KERNEL);
	  }else{
	    i=usb_submit_urb(urb, SLAB_ATOMIC);
	  }
	#else
	 i=usb_submit_urb(urb);
	#endif
	
	DEBUG(3, "<- usb_submit_urb(urb=%p rc=%d)\n",urb,i);
	return i;
}

STATIC inline int xusb_unlink_urb(struct urb *urb) {
int i;
        DEBUG(3, "-> usb_unlink_urb(urb=%p)\n", urb);
        i=usb_unlink_urb(urb);
	DEBUG(3, "<- usb_unlink_urb(urb=%p rc=%d)\n",urb,i);
	return i;
}

STATIC void submit_ctrl(usb_cardman_t *cardman, 
			unsigned char requesttype, 
			unsigned char request, 
			unsigned short wValue, unsigned short wIndex ) {

	DEBUG(3, "-> submit_ctrl\n");

	cardman->ctrl->transfer_buffer=NULL;
	cardman->ctrl->transfer_buffer_length=0;
	cardman->ctrl->start_frame=0;
	cardman->irq->start_frame=-1;
        #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
          cardman->dr->bRequestType=requesttype;
          cardman->dr->bRequest=request;
          cardman->dr->wValue=wValue;
          cardman->dr->wIndex=wIndex;
          cardman->dr->wLength=0;
	#else
          cardman->dr->requesttype=requesttype;
          cardman->dr->request=request;
          cardman->dr->value=wValue;
          cardman->dr->index=wIndex;
          cardman->dr->length=0;
	#endif
	cardman->ctrl->dev=cardman->dev;
	xusb_submit_urb(cardman->ctrl);
	
	DEBUG(3, "<- submit_ctrl\n");
}

STATIC inline void cardman_delete (struct usb_cardman *cardman) {
	int i;

	DEBUG(3, "-> cardman_delete\n");
  
	if (cardman->irq != NULL) {
		usb_free_urb(cardman->irq);
		DEBUG(5, "freeing irq urb\n");
	}
	if (cardman->ctrl != NULL){
		usb_free_urb(cardman->ctrl);
		DEBUG(5, "freeing ctrl urb\n");
	}
	if (cardman->rctl != NULL){
		usb_free_urb(cardman->rctl);
		DEBUG(5, "freeing rctrl urb\n");
	}
	if (cardman->ibuf != NULL){
		kfree(cardman->ibuf);
		DEBUG(5, "freeing ibuf\n");
	}
	if (cardman->dr != NULL){
		kfree(cardman->dr);
		DEBUG(5, "freeing dr\n");
	} 

	i=0;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	#else
       	 while ((i<CM2020_MAX_DEV) && (cardman_table[i] != cardman)) i++;
	 if (i<CM2020_MAX_DEV)  
	 	cardman_table[i]=NULL;
	#endif
  
	kfree(cardman);
	DEBUG(3, "<- cardman_delete\n");
}

#ifdef  ATR_DEBUG
STATIC void dumpatr(usb_cardman_t *cardman) {
unsigned long flags;
int i;
	// save_flags(flags);
	// sti();
	DEBUG(1, "ATR: ");
	for(i=0;i<cardman->atr_len;i++){
	  if (i%8==0) DEBUG(1, "\n");
	  DEBUG(1, "%.2x ",cardman->atr[i]);
	}
	DEBUG(1, "\n");
	// restore_flags(flags);
}
#endif

STATIC int parse_atr(usb_cardman_t *cardman) {
unsigned char anyT0,anyT1,ta1;
unsigned char ch;
unsigned char ix,ifno,done;
unsigned char i;

	DEBUG(3, "-> enter parse_atr\n");
#ifdef ATR_DEBUG
	dumpatr(cardman);
#endif //ATR_DEBUG
	if (cardman->atr_len<1) {
		#ifdef ATR_DEBUG
	          DEBUG(5,"need at least 1 byte of atr\n");
		#endif // ATR_DEBUG
		goto return_0;
	}
	clear_bit(IS_ANY_T0, (void *)&cardman->flags);
	clear_bit(IS_ANY_T1, (void *)&cardman->flags);
	clear_bit(T1MODE2, (void *)&cardman->flags);
	cardman->bIFSD=0;
	cardman->bIFSC=0;

	if (cardman->atr[0]==0x03)
	  str_invert_revert(cardman->atr, cardman->atr_len);
	if (cardman->atr[0]==0x3f) 
	  set_bit(IS_INVERSE, (void *)&cardman->flags);
	else 
	  clear_bit(IS_INVERSE, (void *)&cardman->flags);
	if (cardman->atr_len<3) {
                #ifdef ATR_DEBUG
	 	  DEBUG(5,"need at least 3 bytes of atr\n");
                #endif //ATR_DEBUG
		goto return_0;
	}
	ix=1;
	ifno=1;
	ch=cardman->atr[1];
	cardman->proto=0;
	anyT1=anyT0=done=0;
	ta1=0x11; /* defaults to 9600 */
	
	do {
	        if ((ifno==1) && (ch&0x10)){ //TA(1)
		  /* alternate baudrate */
		  ta1=cardman->atr[2];
		  DEBUG(4,"parse_atr: Set alternate baudrate to 0x%.2x\n", ta1);
		}
	        else if ((ifno==2) && (ch&0x10)){ //TA(2) specific mode
		  DEBUG(4, "Card indicates specific mode!\n");
		  ta1=0x11;
		}
		
		DEBUG(4,"parse_atr: in loop ta1 = 0x%.2x\n", ta1);
		#ifdef ATR_DEBUG
		  DEBUG(5,"ch=%.2x\n",ch);
		#endif //ATR_DEBUG
		
		if ((ifno>2) && (anyT1==1) && (ch&0x10) && (ix+1<cardman->atr_len)){
		  cardman->bIFSC=cardman->atr[ix+1];
		  DEBUG(4, "cardman->bIFSC= %i\n", cardman->bIFSC);
		}

		ix+=((ch&0x10)>>4)		/* no of int.face chars */
			+ ((ch&0x20)>>5)
			+ ((ch&0x40)>>6)
			+ ((ch&0x80)>>7);
		#ifdef ATR_DEBUG
		  DEBUG(5,"ix=%d\n",ix);
                #endif //ATR_DEBUG

		if (ix>=cardman->atr_len || ix>=MAX_ATR) {
			#ifdef ATR_DEBUG
		          DEBUG(5, "not enough atr-bytes\n");
                        #endif // ATR_DEBUG
			goto return_0;
		}

		ifno++;
		if (ch&0x80) {  /* TDi */
		        #ifdef ATR_DEBUG
			  DEBUG(5,"TDi present\n");
			#endif //ATR_DEBUG
			ch=cardman->atr[ix];
			if (ch&0x0f){
			  DEBUG(5, "card is capable of T=1\n");
				anyT1=1;
			}else{
			  DEBUG(5, "card is capable of T=0\n");
				anyT0=1;
			}
		}
		else done=1;
	} while(!done);

	#ifdef ATR_DEBUG
	 DEBUG(5,"parse_atr: ix=%d noHist=%d anyT1=%d anyT0=%d\n",
	 	ix,cardman->atr[1]&15,anyT1, anyT0);
	#endif //ATR_DEBUG

	if (ix+1+(cardman->atr[1]&0x0f)+anyT1!=cardman->atr_len) {
		DEBUG(5,"parse_atr: length error\n");
		goto return_0;
	}
	if (anyT0)
		set_bit(IS_ANY_T0, (void *)&cardman->flags);
	if (anyT1) {	/* compute csum */
		cardman->atr_csum=0;
		for(i=1;i<cardman->atr_len;i++)
			cardman->atr_csum^=cardman->atr[i];
		if (cardman->atr_csum) {
			set_bit(IS_BAD_CSUM, (void *)&cardman->flags);
			#ifdef ATR_DEBUG
			  DEBUG(5,"parse_atr: bad checksum\n");
			#endif //ATR_DEBUG
			goto return_0;
		}
		if (anyT0==0)
			cardman->proto=1;
		set_bit(IS_ANY_T1, (void *)&cardman->flags);
	}
	DEBUG(4,"parse_atr: before asignment ta1 = 0x%.2x\n", ta1);
	cardman->ta1=ta1;
	DEBUG(4, "parse_atr: cardman->ta1 = 0x%.2x\n", cardman->ta1);


	// Handle  special cards
	if ((cardman->atr_len == 24  && 
	     (cardman->atr[0] == 0x3B || cardman->atr[0] == 0x3F) &&
	     cardman->atr[1] == 0xEF &&
	     cardman->atr[2] == 0x00 &&
	     cardman->atr[3] == 0xFF &&
	     cardman->atr[4] == 0x81 &&
	     cardman->atr[5] == 0x31 &&
	     cardman->atr[7] == 0x45 &&
	     cardman->atr[8] == 0x65 &&
	     cardman->atr[9] == 0x63) ||
	    (cardman->atr_len == 25 &&
	     (cardman->atr[0] == 0x3B  || cardman->atr[0] == 0x3F) &&
	     cardman->atr[1] == 0xFF &&
	     cardman->atr[3] == 0x00 &&
	     cardman->atr[4] == 0xFF &&
	     cardman->atr[5] == 0x81 &&
	     cardman->atr[6] == 0x31 &&
	     cardman->atr[8] == 0x45 &&
	     cardman->atr[9] == 0x65 &&
	     cardman->atr[10] == 0x63) ||
	    ((cardman->atr[0]== 0x3B ) &&
	     (cardman->atr[1] & 0x0F) == 0x0F &&
	     cardman->atr[cardman->atr_len - (cardman->atr[1] & 0x0F) - 1 ] == 0x45   &&
	     cardman->atr[cardman->atr_len - (cardman->atr[1] & 0x0F)     ] == 0x50   &&
	     cardman->atr[cardman->atr_len - (cardman->atr[1] & 0x0F) + 1 ] == 0x41   &&
	     cardman->atr[cardman->atr_len - (cardman->atr[1] & 0x0F) + 2 ] == 0x00)
	    )
	{
	  DEBUG(4, "German ec-Card detected!\n");
	  set_bit(T1MODE2, (void*)&cardman->flags);
	  cardman->bIFSD=252;
	  cardman->ta1=0x11;
	}else if(cardman->atr_len == 24 &&
		 cardman->atr[0]  == 0x3B &&
		 cardman->atr[1]  == 0xBF &&
		 cardman->atr[2]  == 0x18 &&
		 cardman->atr[3]  == 0x00 &&
		 cardman->atr[4]  == 0x81 &&
		 cardman->atr[5]  == 0x31 &&
		 cardman->atr[6]  == 0xFE &&
		 cardman->atr[8]  == 0x01 &&
		 cardman->atr[9]  == 0x0D &&
		 cardman->atr[10] == 0x44 &&
		 cardman->atr[11] == 0x4D &&
		 cardman->atr[12] == 0x50 &&
		 cardman->atr[13] == 0x53 &&
		 cardman->atr[14] == 0x41 &&
		 cardman->atr[15] == 0x4D)
	{
	  DEBUG(4, "PSAM V24 detected!\n");
	  set_bit(T1MODE2, (void*)&cardman->flags);
	  cardman->bIFSD=252;
	  cardman->ta1=0x11;
	}else if (cardman->atr_len == 21 &&
		  cardman->atr[0]  == 0x3B &&
		  cardman->atr[1]  == 0xBC &&
		  cardman->atr[2]  == 0x18 &&
		  cardman->atr[3]  == 0x00 &&
		  cardman->atr[4]  == 0x81 &&
		  cardman->atr[5]  == 0x31 &&
		  cardman->atr[6]  == 0xFE &&
		  cardman->atr[7]  == 0x64 &&
		  cardman->atr[8]  == 0x01 &&
		  cardman->atr[9]  == 0x0A &&
		  cardman->atr[16] == 0x1e &&
		  cardman->atr[17] == 0x1c)
	{
	  DEBUG(4, "PSAM V30 detected!\n");
	  set_bit(T1MODE2, (void*)&cardman->flags);
	  cardman->bIFSD=254;
	  cardman->ta1=0x11;
	}else if (cardman->atr_len == 15 &&
		  cardman->atr[0] == 0x3B &&
		  cardman->atr[1] == 0xE6 &&
		  cardman->atr[2] == 0x00 &&
		  cardman->atr[3] == 0xFF &&
		  cardman->atr[4] == 0x81 &&
		  cardman->atr[5] == 0x31 &&
		  cardman->atr[6] == 0x42 &&
		  cardman->atr[7] == 0x45)
	{
	  DEBUG(4, "Moneo card detected!\n");
	  set_bit(T1MODE2, (void*)&cardman->flags);
	  cardman->bIFSD=254;
	  cardman->ta1=0x11;
	}else if (cardman->atr_len == 25 &&
		  cardman->atr[0] == 0x3B  &&
		  //cardman->atr[1] == 0xe9  &&
		  cardman->atr[2] == 0x18 &&
		  cardman->atr[3] == 0x00 &&
		  cardman->atr[4] == 0xff &&
		  cardman->atr[5] == 0x81 &&
		  cardman->atr[6] == 0x31 &&
		  cardman->atr[7] == 0xFE &&
		  cardman->atr[8] == 0x45 &&
		  cardman->atr[11] != '@') //excl. Smart@Cafe
	{
	  DEBUG(4, "Secartis card detected!\n");
	  set_bit(T1MODE2, (void*)&cardman->flags);
	  cardman->bIFSD=252;
	  cardman->ta1=0x11;
	}else{
	  clear_bit(T1MODE2, (void*)&cardman->flags);
	  // Signal t1lib to use default values
	  cardman->bIFSD=0;
	  cardman->bIFSC=0;
	}
	if (test_bit(T1MODE2, (void*)&cardman->flags)>0){
	  DEBUG(4,"T1Mode2 enabled, bIFSD = %i\n", cardman->bIFSD);
	}

	DEBUG(3, "<- parse_atr (rc = 1 (okay))\n");
	return 1;
return_0:
	DEBUG(3, "<- parse_atr (rc = 0 (error))\n");
	return 0;
}

STATIC void set_card_parameter(usb_cardman_t *,unsigned char, unsigned char, unsigned char);
STATIC int  set_protocol(usb_cardman_t *, ptsreq_t *);
STATIC void request_status(usb_cardman_t *);
STATIC void request_atr(usb_cardman_t *, int);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC void cardman_ctrl_callback(struct urb *, struct pt_regs*);
STATIC void cardman_rctl_callback(struct urb *, struct pt_regs*);
STATIC void cardman_irq_callback(struct urb *, struct pt_regs*);
#else
STATIC void cardman_ctrl_callback(struct urb *);
STATIC void cardman_rctl_callback(struct urb *);
STATIC void cardman_irq_callback(struct urb *);
#endif

STATIC inline void cmwaitforOpFinish (usb_cardman_t *cardman){
  while(test_bit(IS_OP_PENDING, (void *)&cardman->flags)) {
    DEBUG(5,"cardman: wait for operation to finish: go to sleep\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(5,"wait_event IS_OP_PENDING\n");
  wait_event_interruptible(cardman->waitq,!test_bit(IS_OP_PENDING, (void *)&cardman->flags));
#else
    interruptible_sleep_on(&cardman->waitq);
#endif
  }
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  if (cardman->irq->status == -EPIPE){
    DEBUG(5, "Clearing stalled IRQ pipe and resubmitting IRQ URB\n");
    usb_clear_halt(cardman->dev, cardman->irq->pipe);
    cardman->irq->start_frame=-1;
    xusb_submit_urb(cardman->irq);
  }
  #endif
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC void cardman_ctrl_callback(struct urb *urb, struct pt_regs* pt_reg) {
#else
STATIC void cardman_ctrl_callback(struct urb *urb) {
#endif
usb_cardman_t *cardman=(usb_cardman_t *)urb->context;
#ifdef CM2020_DEBUG
int i;
 DEBUG(4,"-> ctrl callback urb=%p status=%d frame=%d len=%d data=",
       urb,urb->status,urb->start_frame,urb->actual_length);
 for(i=0;i<urb->actual_length;i++){
   if (i%8==0) DEBUG(4,"\n");
   DEBUG(4,"%.2x ",((unsigned char *)urb->transfer_buffer )[i]);
 }
 DEBUG(4,"\n");
#endif //CM2020_DEBUG
 switch (cardman->op){
 case CB_SET_PARAMETER:
   DEBUG(4,"ctrl_callback: waking for set_parameter\n");
   if (urb->status != 0){ 
     cardman->op=CB_CARD_OFF;
     submit_ctrl(cardman,CM2020_REQT_WRITE,CM2020_CARD_OFF, 0, 0);
   }else{
     cardman->op = CB_READ_STATUS;
     cardman->ttl=cardman->ttl_hi=2;
     request_status(cardman);
   }
   break;
 case CB_CARD_OFF:
   DEBUG(4,"ctrl_callback: waking for card_off\n");
   cardman->op=CB_READ_STATUS;
   cardman->ttl=cardman->ttl_hi=10;
   request_status(cardman);
   break;
 case CB_WRITE_PTS:
   DEBUG(4,"ctrl_callback: waking for write_pts\n");
   break;
 case CB_READ_STATUS:
   DEBUG(4, "ctrl_callback: waking for request_status\n");
   break;
 case CB_WRITE_T1:
   DEBUG(4, "ctrl_callback: now watch T=1 communication progress\n");
   cardman->op=CB_PROG_T1;
   break;
 default:
   DEBUG(4,"ctrl_callback: default action (do nothing)\n");
   break;
 }
 DEBUG(2,"<- ctrl callback\n");
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC void cardman_rctl_callback(struct urb *urb, struct pt_regs* pt_reg) {
#else
STATIC void cardman_rctl_callback(struct urb *urb) {
#endif
  usb_cardman_t *cardman=(usb_cardman_t *)urb->context;

#ifdef CM2020_DEBUG
int i;
 DEBUG(2,"-> rctl callback urb=%p status=%d frame=%d len=%d data=",
       urb,urb->status,urb->start_frame,urb->actual_length);
 if (urb->actual_length){
   for(i=0;i<urb->actual_length;i++){
     if (i%8==0) DEBUG(4,"\n");
     DEBUG(4,"%.2x ",cardman->rcbuf[i]);
   }
 }
 DEBUG(4,"\n");
 #endif //CM2020_DEBUG

 if (urb->status==0 && urb->actual_length) {
   DEBUG(4,"rlen=%d t1_reply=%d\n",cardman->rlen,cardman->t1_reply_len);
   if (cardman->rlen+urb->actual_length<=cardman->t1_reply_len) {
     memcpy(&cardman->rbuf[cardman->rlen],cardman->rcbuf,urb->actual_length);
     cardman->rlen+=urb->actual_length;
     DEBUG(4,"rlen=%d t1_reply=%d\n",cardman->rlen,cardman->t1_reply_len);
     if (cardman->rlen==cardman->t1_reply_len) {
       // wake_up_interruptible(&cardman->waitq);
       cardman->op=CB_READ_STATUS;
       cardman->ttl=cardman->ttl_hi=2;
       DEBUG(4,"request status from rctl_callback\n");
       request_status(cardman);
     }
   }
 }

#if 0
 if (urb->status==-EPIPE) {
   DEBUG(4,"clearing halt (rctl)\n");
   usb_clear_halt(cardman->dev,usb_sndctrlpipe(cardman->dev,0));
 }
#endif

 switch (cardman->op){
 case CB_READ_PTS:
   {
     unsigned char bBaudRate, bStopBits;
     
     bBaudRate = CM2020_FREQUENCY_3_72MHZ + CM2020_BAUDRATE_9600;
     bStopBits = 0x02;
     
     DEBUG(4,"rctl_callback: op=CB_READ_PTS\n");
     if ((urb->actual_length >= 3) || (urb->actual_length <= 4)){
       DEBUG(4,"rctl_callback: PTS reply: length == %d\n", urb->actual_length);
       if ((cardman->pts[0] == cardman->rcbuf[0]) &&
	   (cardman->pts[1] == cardman->rcbuf[1]) &&
	   (cardman->pts[2] == cardman->rcbuf[2]) &&
	   (cardman->pts[3] == cardman->rcbuf[3]))
	 {
	   DEBUG(4,"rctl_callback: PTS reply valid: Calc baudrate and call set_card_parameter()\n");
	   // set baudrate
	   switch (cardman->pts[2]){
	   case 0x11: bBaudRate = CM2020_FREQUENCY_3_72MHZ + CM2020_BAUDRATE_9600; break;
	   case 0x12: bBaudRate = CM2020_FREQUENCY_3_72MHZ + CM2020_BAUDRATE_19200; break;
	   case 0x13: bBaudRate = CM2020_FREQUENCY_3_72MHZ + CM2020_BAUDRATE_38400; break;
	   case 0x14: bBaudRate = CM2020_FREQUENCY_3_72MHZ + CM2020_BAUDRATE_76800; break;
	   case 0x18: bBaudRate = CM2020_FREQUENCY_3_72MHZ + CM2020_BAUDRATE_115200; break;
	   case 0x91: bBaudRate = CM2020_FREQUENCY_5_12MHZ + CM2020_BAUDRATE_9600; break;
	   case 0x92: bBaudRate = CM2020_FREQUENCY_5_12MHZ + CM2020_BAUDRATE_19200; break;
	   case 0x93: bBaudRate = CM2020_FREQUENCY_5_12MHZ + CM2020_BAUDRATE_38400; break;
	   case 0x94: bBaudRate = CM2020_FREQUENCY_5_12MHZ + CM2020_BAUDRATE_76800; break;
	   case 0x98: bBaudRate = CM2020_FREQUENCY_5_12MHZ + CM2020_BAUDRATE_115200; break;
	   }
	   if (test_bit(IS_INVERSE, (void *)&cardman->flags)){
	     bBaudRate |= CM2020_ODD_PARITY;
	     DEBUG(4,"rctl_callback: Apply ODD_PARITY convention!\n");
	   }
	   // set stopBits
	   bStopBits = 2;
	   // Apply
	   set_card_parameter(cardman, CM2020_CARD_ASYNC, bBaudRate, bStopBits);
	   
	 }else if ((cardman->pts[0] == cardman->rcbuf[0]) &&
                   ((cardman->pts[1]&0xef) == cardman->rcbuf[1]) && // b5 indicates PPS1 not present; parameter t are echoed
		   ((cardman->rcbuf[0]^cardman->rcbuf[1]) == cardman->rcbuf[2])
                  )
	 {
	   DEBUG(4, "rctl_callback: Short PPS answer received\n");
	   if (test_bit(IS_INVERSE, (void *)&cardman->flags)){
	     bBaudRate |= CM2020_ODD_PARITY;
	     DEBUG(4,"rctl_callback: Apply ODD_PARITY convention!\n");
	   }
	   // set stopBits
	   bStopBits = 2;
	   // Apply
	   set_card_parameter(cardman, CM2020_CARD_ASYNC, bBaudRate, bStopBits);
	 }else{
	   DEBUG(4,"rctl_callback: PTS reply invalid: request and reply differ\n");
	   cardman->op=CB_CARD_OFF;
	   submit_ctrl(cardman,CM2020_REQT_WRITE,CM2020_CARD_OFF, 0, 0); 
	 }
     }else{
       DEBUG(4,"rctl_callback: PTS reply invalid: Wrong length\n");
       cardman->op=CB_CARD_OFF;
       submit_ctrl(cardman,CM2020_REQT_WRITE,CM2020_CARD_OFF, 0, 0);       
     }
   }	      
   break;
 default: 
   DEBUG(4, "rctl_callback: default action (do nothing)\n");
   break;
 }
 
 DEBUG(2,"<- rctl_callback\n");
}

STATIC inline int cmresubmitirqurb(usb_cardman_t *cardman){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  //resubmit IRQ pipe urb
  DEBUG(4, "Resubmitting irq URB\n");
  return xusb_submit_urb(cardman->irq);
#else
  return 0;
#endif
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC void cardman_irq_callback(struct urb *urb, struct pt_regs* pt_reg) {
#else
STATIC void cardman_irq_callback(struct urb *urb) {
#endif
  usb_cardman_t *cardman=urb->context;
  unsigned char procByte;
  int i;

#ifdef CM2020_DEBUG
  DEBUG(4,"->irq callback urb=%p status=%d frame=%d len=%d interval=%i data=",
	urb,urb->status,urb->start_frame,urb->actual_length, cardman->irq->interval);
  if (urb->status != 0)
    DEBUG(4,"irq_callback: BAD IRQ pipe (%i)\n", urb->status);
  if (cardman->op == CB_T1MODE2)
    DEBUG(4, "irq_callback: op is T1MODE2\n");
  if (urb->actual_length){
    for(i=0;i<urb->actual_length;i++){
      if (i%8==0) DEBUG(4,"\n");
      DEBUG(4,"%.2x ",((unsigned char *)urb->transfer_buffer)[i]);
    }
  }
  DEBUG(4,"\n");
  
#endif
  
  DEBUG(2,"irq_callback: op=%d\n",cardman->op);
  
  if (urb->status==-EPIPE) {	/* stalled endpoint */
    DEBUG(4,KERN_INFO "irq_callback: STALLED (-EPIPE)\n");    
    cardman->op=CB_READ_STATUS;
    cardman->ttl=cardman->ttl_hi=2;
    cardman->t1_reply_len=-1;
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    clear_bit(IS_OP_PENDING, (void*)&cardman->flags);
    DEBUG(2,"wake up IS_OP_PENDING\n");
    wake_up_interruptible(&cardman->waitq);
#else
    request_status(urb->context);
#endif
    DEBUG(2,"<- irq callback (stalled endpoint)\n");
    return;
  }

  urb->start_frame++;
  #ifdef DEBUG2
  if (urb->actual_length)
    {
      int i;
      printk("urblen=%d\n",urb->actual_length);
      for(i=0;i<urb->actual_length;i++)
	printk("%.2x ",((unsigned char *)urb->transfer_buffer)[i]);
      printk("\n");
    }
  #endif //DEBUG2
  if (urb->actual_length==0) {
    if (urb->start_frame>0) {
      cardman->ttl--;
      DEBUG(4,"irq_callback: ttl=%d rlen=%d\n",cardman->ttl,cardman->rlen);
      if (cardman->ttl<=0) {	/* timeout */
	if (cardman->op == CB_T1MODE2){
	  cardman->bwt=-1;
	  cardman->t1_reply_len=-1;
	}
      	if (cardman->op != CB_READ_STATUS){
	  DEBUG(4,"irq_callback: ttl<=0, requesting status\n");
	  cardman->op=CB_READ_STATUS;
	  cardman->ttl=cardman->ttl_hi=2;
	  if (cardman->rlen==0)
	    cardman->t1_reply_len=-1;
  	  else
	    cardman->t1_reply_len=0;
	  request_status(urb->context);
	}else{
	  DEBUG(4, "irq_callback: Read status timeout reached (unlock OP_PENDING)\n");
	  cardman->op=CB_NOP;
	  clear_bit(IS_ATR_VALID, (void *)&cardman->flags);
	  clear_bit(IS_ATR_PRESENT, (void *)&cardman->flags);
	  clear_bit(IS_INVERSE, (void *)&cardman->flags);
          clear_bit(IS_OP_PENDING, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up event is unknown\n");
    wake_up_interruptible(&cardman->waitq);
#else
          wake_up_interruptible(&cardman->waitq);
#endif
	}
      }
    }
   
    DEBUG(4, "<- irq_callback (no data)\n");
    goto finish_irq_callback;
  }
  /* cardman->ttl therefore acts as a character waiting timeout */
  cardman->ttl=cardman->ttl_hi;	/* reset timeout */
  
  switch(cardman->op) {
  case CB_NOP:
    DEBUG(4,"irq_callback: op=CB_NOP?\n");
    clear_bit(IS_OP_PENDING, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up IS_OP_PENDING\n");
    wake_up_interruptible(&cardman->waitq);
#else
    wake_up_interruptible(&cardman->waitq);
#endif
    break;
  case CB_READ_STATUS:
    DEBUG(4,"irq_callback: op=CB_READ_STATUS, buf=%.2x\n",cardman->ibuf[0]);
    DEBUG(4,"irq_callback: ibuf=%p transfer_buffer=%p\n",
	  cardman->ibuf,urb->transfer_buffer);
    if (urb->actual_length!=1){
      DEBUG(4, "irq_callback: got no status byte\n");
    }else{
   
      cardman->card_state=(cardman->ibuf[0] | ((cardman->ibuf[0] & 0x02) << 6)) & CM2020_STATUS_MASK; /* cm6020 firmware bug */
    
      DEBUG(4,"irq_callback: cardman->card_state=%.2x\n",cardman->card_state);
      if (cardman->card_state==0) {	/* no card */
	DEBUG(4,"irq_callback: no card inserted\n");
	cardman->op=CB_NOP;
	cardman->atr_len=0;
	clear_bit(IS_ATR_PRESENT, (void *)&cardman->flags);
	clear_bit(IS_ATR_VALID, (void *)&cardman->flags);
	clear_bit(IS_INVERSE, (void *)&cardman->flags);
	clear_bit(IS_OP_PENDING, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up event is unknown\n");
	wake_up_interruptible(&cardman->waitq);
#else
	wake_up_interruptible(&cardman->waitq);
#endif
      }
      /* card inserted, but not powered: power on and get atr */
      else if (cardman->card_state==0x40) {
	DEBUG(4, "irq_callback: card inserted but not powered\n");
	
	clear_bit(IS_ATR_PRESENT, (void *)&cardman->flags);
	clear_bit(IS_ATR_VALID, (void *)&cardman->flags);
	clear_bit(IS_INVERSE, (void *)&cardman->flags);
	clear_bit(IS_OP_PENDING, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up event is unknown\n");
	wake_up_interruptible(&cardman->waitq);      
#else
	wake_up_interruptible(&cardman->waitq);      
#endif
      }
      /* card inserted and powered: we've read the atr before */
      else if (cardman->card_state==0xc0) {
	DEBUG(4,"irq_callback: card powered\n");
	cardman->op=CB_NOP;
	clear_bit(IS_OP_PENDING, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up IS_OP_PENDING\n");
	wake_up_interruptible(&cardman->waitq);
#else
	wake_up_interruptible(&cardman->waitq);
#endif
      }
    }
    break;
  case CB_READ_ATR:
    DEBUG(4,"irq_callback: op=CB_READ_ATR\n");
    if (urb->actual_length+cardman->atr_len<MAX_ATR && urb->actual_length) {
      DEBUG(4, "irq_callback: atr_len=%d\n",cardman->atr_len);
      DEBUG(4, "irq_callback: urb_len=%d\n",urb->actual_length);
      DEBUG(4, "irq_callback: inverse=%d\n",test_bit(IS_INVERSE, (void *)&cardman->flags));
      if (test_bit(IS_INVERSE, (void *)&cardman->flags))
	str_invert_revert(urb->transfer_buffer,urb->actual_length);
#if 0
      memcpy(cardman->atr+cardman->atr_len,cardman->ibuf,
	     urb->actual_length);
#endif
      memcpy(cardman->atr+cardman->atr_len,urb->transfer_buffer,
	     urb->actual_length);
      cardman->atr_len+=urb->actual_length;
      if (parse_atr(cardman)) {	/* atr okay */
#ifdef CM2020_DEBUG
	DEBUG(4,"irq_callback: ATR: ");
	for(i=0;i<cardman->atr_len;i++)
	  DEBUG(4,"%.2x ",cardman->atr[i]);
	DEBUG(4,"\n");
	DEBUG(4,"proto=%d inverse=%d\n",cardman->proto,test_bit(IS_INVERSE, (void *)&cardman->flags));
#endif
	DEBUG(4,"irq_callback: atr okay, setting ATR_VALID/ATR_PRESENT\n");
	set_bit(IS_ATR_VALID, (void *)&cardman->flags);
	set_bit(IS_ATR_PRESENT, (void *)&cardman->flags);
	cardman->ttl=0;	/* atr ready, abort timeout */

      }
      if (cardman->atr_len>0 && !(cardman->atr[0]==0x3b || cardman->atr[0]==0x3f)) {
	/* invalid atr */
	DEBUG(4,"irq_callback: ATR invalid\n");
	clear_bit(IS_INVERSE, (void *)&cardman->flags);
	clear_bit(IS_ATR_VALID, (void *)&cardman->flags);
	set_bit(IS_ATR_PRESENT, (void *)&cardman->flags);
	cardman->ttl=0;	/* atr ready, abort timeout */
	cardman->irq->start_frame=-1;
	cardman->op=CB_SET_PARAMETER;
	DEBUG(4,"irq_callback: set_card_parameter to default!\n");
	set_card_parameter(cardman,CM2020_CARD_ASYNC,CM2020_BAUDRATE_9600, 0x02);
      }
    }
    break;
  case CB_WRITE_PTS:
    DEBUG(4,"irq_callback: op=CB_WRITE_PTS\n");
    DEBUG(4,"irq_callback: URB length = %i, ibuf[0] = 0x%.2x\n",urb->actual_length, cardman->ibuf[0]);
    if (urb->actual_length==2 && ((cardman->ibuf[0]>=3) && (cardman->ibuf[0]<=4))) {
      DEBUG(4,"irq_callback: PTS complete!\n");
      cardman->op=CB_READ_PTS;
      cardman->rctl->transfer_buffer_length=cardman->ibuf[0];
      cardman->ttl=cardman->ttl_hi=120;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
      cardman->dr->bRequestType=CM2020_REQT_READ;
      cardman->dr->bRequest=0;
      cardman->dr->wValue=0;
      cardman->dr->wIndex=0;
      cardman->dr->wLength=cardman->rctl->transfer_buffer_length;
#else
      cardman->dr->requesttype=CM2020_REQT_READ;
      cardman->dr->request=0;
      cardman->dr->value=0;
      cardman->dr->index=0;
      cardman->dr->length=cardman->rctl->transfer_buffer_length;
#endif
      cardman->rctl->dev=cardman->dev;
      cardman->irq->start_frame=-1;
      xusb_submit_urb(cardman->rctl);
    }else{
      // time out for waiting for valid number of bytes to be returned
      if (cardman->ptsttl>=0){
	cardman->ptsttl--;
      }else{
	DEBUG(4,"irq_callback: PPS timeout reached!\n");
	cardman->op=CB_CARD_OFF;
	submit_ctrl(cardman,CM2020_REQT_WRITE,CM2020_CARD_OFF, 0, 0);
	DEBUG(4, "*** CB_CARD_OFF wait 50 msec\n");
	mdelay(50);
	DEBUG(4, "*** CB_CARD_OFF waiting 50 msec done\n");
      }
    }
    break;
  case CB_PROG_T1:
    #ifdef CM2020_DEBUG
  DEBUG(2,"->irq callback urb=%p status=%d frame=%d len=%d interval=%i data=",
	urb,urb->status,urb->start_frame,urb->actual_length, cardman->irq->interval);
  
  if (urb->actual_length){
    for(i=0;i<urb->actual_length;i++){
      if (i%8==0) DEBUG(2,"\n");
      DEBUG(2,"%.2x ",((unsigned char *)urb->transfer_buffer)[i]);
    }
  }
  DEBUG(2,"\n");
  
#endif
    DEBUG(4,"irq_callback: op=CB_PROG_T1\n");
    DEBUG(4,"irq_callback: t1_len: urb->status=%d\n",urb->status);
    DEBUG(4,"irq_callback: cardman->ttl = %i\n", cardman->ttl);
    DEBUG(4,"irq_callback: cardman->bwt = %i\n", cardman->bwt);

    if (cardman->bwt<=0) {
      DEBUG(4,"irq_callback: T1 Timeout (unresponsive card) !\n");
      cardman->op=CB_READ_STATUS;
      cardman->ttl=cardman->ttl_hi=2;
      cardman->t1_reply_len=-1;
      request_status(urb->context);
    }    
    else if (urb->actual_length==2 && cardman->ibuf[0]>=4) {
      if (test_bit(IS_INVERSE, (void*)&cardman->flags)){
	cardman->ibuf[1]=invert_revert(cardman->ibuf[1]);
	DEBUG(4,"irq_callback: INVERSE convention applies\n");
      }
      if (cardman->ibuf[0]==4+cardman->ibuf[1]) {
	DEBUG(4,"irq_callback: CB_PROG_T1: response complete\n");
	DEBUG(4,"irq_callback: CB_PROG_T1: t1_reply_len=%d\n",
	      cardman->ibuf[0]);
	cardman->t1_reply_len=cardman->ibuf[0];
	cardman->rctl->transfer_buffer_length=cardman->t1_reply_len;
	cardman->op=CB_READ_T1;
	cardman->ttl=cardman->ttl_hi=12;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	cardman->dr->bRequestType=CM2020_REQT_READ;
	cardman->dr->bRequest=0;
	cardman->dr->wValue=0;
	cardman->dr->wIndex=0;
	cardman->dr->wLength=cardman->t1_reply_len;
#else
	cardman->dr->requesttype=CM2020_REQT_READ;
	cardman->dr->request=0;
	cardman->dr->value=0;
	cardman->dr->index=0;
	cardman->dr->length=cardman->t1_reply_len;
#endif
	cardman->rctl->dev=cardman->dev;
	cardman->irq->start_frame=-1;
	DEBUG(4,"irq_callback: submit c2\n");
	xusb_submit_urb(cardman->rctl);
      }
    }
    cardman->bwt--;
    break;
  case CB_WRITE_T0:
    once_more:
    DEBUG(4,"irq_callback: op=CB_WRITE_T0\n");
    procByte=cardman->ibuf[0];
    DEBUG(4,"irq_callback: pbyte=%.2x t0ins=%.2x\n",
	  procByte,
	  cardman->t0_ins);
    if (test_bit(IS_INVERSE, (void *)&cardman->flags))
      str_invert_revert(&procByte,1);
    /* NULL */
    if (procByte==0x60) {
      int ii;
      DEBUG(4,"irq_callback: T0 WWT\n");
      if (urb->actual_length>1){ // more data (SW1SW2!?) were sent in the same packet
	for(ii=0; ii<urb->actual_length-1; i++){
	  cardman->ibuf[ii] = cardman->ibuf[ii+1];
	}
	urb->actual_length--;
	DEBUG(4, "irq_callback: once more\n");
	goto once_more;
      }
      break;	/* WWT */
    }
    /* ACK */
    if ((procByte&0xfe)==(cardman->t0_ins&0xfe)) {
      DEBUG(4,"irq_callback: T0 ACK\n");
      i=cardman->t0_data_len;
    }
    /* NACK */
    else if ((~procByte&0xfe)==(cardman->t0_ins&0xfe)) {
      DEBUG(4,"irq_callback: T0 NACK\n");
      i=1;
    }
    else {
      cardman->op=CB_WRITE_T0_SW1SW2;
      DEBUG(4,"irq_callback: T0 SW1SW2\n");
      goto t0_sw1sw2;
    }
    
    /* buffer underflow */
    if (cardman->t0_data_len==0) {
      cardman->op=CB_READ_STATUS;
      cardman->ttl=cardman->ttl_hi=2;
      cardman->t1_reply_len=-1;
      request_status(urb->context);
      break;
    }
    cardman->ctrl->transfer_buffer+=cardman->t0_data_off;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
    cardman->ctrl->transfer_buffer_length
      = cardman->dr->wLength
      = i;
#else
    cardman->ctrl->transfer_buffer_length 
      = cardman->dr->length
      = i;
#endif
    cardman->t0_data_len-=i;
    cardman->t0_data_off=i;
    cardman->irq->start_frame=-1;
    cardman->ctrl->dev=cardman->dev;
    xusb_submit_urb(cardman->ctrl);
    break;
  t0_sw1sw2:
  case CB_WRITE_T0_SW1SW2:
    DEBUG(4,"irq_callback: op=CB_WRITE_t0_SW1SW2\n");
    /* SW1SW2 */
    if (cardman->rlen==0) { /* this is SW1 */
      DEBUG(4,"irq_callback: SW1\n");
      if((procByte>=0x60 && procByte<=0x6f)
	 || (procByte>=0x90 && procByte<=0x9f)) {
	cardman->rbuf[0]=cardman->ibuf[0];
      }
      if (urb->actual_length==1) {	/* SW2 follows*/
	cardman->rlen=1;	/* seperately */
	break;
      }
    }
    /* this can only be SW2 */
    DEBUG(4,"irq_callback: SW2\n");
    cardman->rbuf[1]=cardman->ibuf[urb->actual_length-1];
    cardman->rlen=2;
    cardman->op=CB_READ_STATUS;
    cardman->ttl=cardman->ttl_hi=2;
    cardman->t1_reply_len=0;
    request_status(urb->context);
    break;
  case CB_READ_T0:
    DEBUG(4,"irq_callback: op=CB_READ_T0\n");
    /* ignore the first 8 bytes */
    if (urb->start_frame==0 && urb->actual_length==8)
      cardman->op=CB_READ_T0_DATA;
    break;
  case CB_READ_T0_DATA:
    DEBUG(4,"irq_callback: op=CB_READ_T0_DATA\n");
    /* ignore the first 8 bytes */
    // if (urb->start_frame==0 && urb->actual_length==8)
    // 	break;
#ifdef DEBUG2
    printk("irq_callback: expect=%d rlen=%d urblen=%d\n",
	   cardman->t0_expected_reply_len,
	   cardman->rlen,
	   urb->actual_length);
#endif
    memcpy(&cardman->rbuf[cardman->rlen],cardman->ibuf,urb->actual_length);
    cardman->rlen+=urb->actual_length;
    DEBUG(4,"irq_callback: expect=%d rlen=%d  urb->actual_length=%d\n",
	  cardman->t0_expected_reply_len,
	  cardman->rlen,
	  urb->actual_length);
    /* we received the complete answer */
    if (cardman->rlen==cardman->t0_expected_reply_len) {
      cardman->op=CB_READ_STATUS;
      cardman->ttl=cardman->ttl_hi=2;
      cardman->t1_reply_len=0;
      request_status(urb->context);
    }
    
    if (cardman->rlen==2 && cardman->t0_expected_reply_len>2 && (cardman->rbuf[0]&0xf0)==0x60){
      int tm = 100/cardman->bInterval;
      if (tm<4)
	tm=4;
      DEBUG(4, "irq_callback: reducing timeout because of an error when reading T=0  data\n");
      if (cardman->ttl > tm)
	cardman->ttl = tm;
    }
    break;
 case CB_T1MODE2:
   DEBUG(4, "irq_callback: op=CB_T1MODE2\n");
   DEBUG(4, "irq_callback: t1_reply_len= %i, rlen= %i,actual_length = %i\n", cardman->t1_reply_len, cardman->rlen, urb->actual_length);
   memcpy(&cardman->rbuf[cardman->rlen],cardman->ibuf,urb->actual_length);
   cardman->rlen+=urb->actual_length;
   if ((cardman->t1_reply_len==0) && (cardman->rlen >= 3)){
     if (test_bit(IS_INVERSE, (void*)&cardman->flags)){
       cardman->t1_reply_len=4+invert_revert(cardman->rbuf[2]);
     }else{
       cardman->t1_reply_len=4+cardman->rbuf[2];
     }
   } 
   if ((cardman->t1_reply_len != 0) && (cardman->rlen==cardman->t1_reply_len)){
     cardman->op=CB_READ_STATUS;
     cardman->ttl=cardman->ttl_hi=2;
     cardman->t1_reply_len=0;
     request_status(urb->context);
   }
   break;
  }

 finish_irq_callback:
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    cmresubmitirqurb(cardman);
  #endif

  DEBUG(2,"<- cardman_irq_callback\n");
  return;
}

#if 0
STATIC void cmu_ctrl_callback(struct urb *urb) {
usb_cardman_t *cardman=(usb_cardman_t *)urb->context;

#ifdef CM2020_DEBUG
unsigned long flags;
int i;
	DEBUG(2,"-> ctrl callback urb=%p status=%d frame=%d len=%d data=",
		urb,urb->status,urb->start_frame,urb->actual_length);
	if (urb->actual_length){
          for(i=0;i<urb->actual_length;i++){
	    if (i%8==0) DEBUG(4,"\n");
            DEBUG(4,"%.2x ",((char *)urb->transfer_buffer)[i]);
	  }
        }
	DEBUG(4,"\n");
#endif

	/* stalled endpoint */
	if (urb->status==-EPIPE || urb->status==-EILSEQ) {
		// DEBUG(4,KERN_INFO "STALLED,  requesting status\n");
		cardman->op=CB_READ_STATUS;
		cardman->ttl=cardman->ttl_hi=2;
		request_status(urb->context);
		return;
	}

	if (cardman->op == CB_WRITE_T1){
	  DEBUG(4, "CB_WRITE_T1: now watch T=1 communication progress, new status is CB_PROG_T1\n");
	  cardman->op=CB_PROG_T1;
	  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	  xusb_submit_urb(cardman->irq);
	  #endif
	}

	DEBUG(2, "<- ctrl_callback\n");
}    
#endif

STATIC void set_card_parameter(usb_cardman_t *cardman,unsigned char bCardType, unsigned char bBaudRate, unsigned char bStopBits) {
	DEBUG(3,"->set_card_parameter\n");
	
	cardman->op=CB_SET_PARAMETER;
	
	cardman->ctrl->transfer_buffer=NULL;
	cardman->ctrl->transfer_buffer_length=0;
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
          cardman->dr->bRequestType=CM2020_REQT_WRITE;
          cardman->dr->bRequest=CM2020_SET_PARAMETER;
	  cardman->dr->wValue=(bBaudRate<<8) | bCardType;
	  DEBUG(5,"dr->value(BaudRate, Cardtype) =%.4x\n",cardman->dr->wValue);
          cardman->dr->wIndex=bStopBits;
          cardman->dr->wLength  =0;
	#else
          cardman->dr->requesttype=CM2020_REQT_WRITE;
          cardman->dr->request=CM2020_SET_PARAMETER;
	  cardman->dr->value=(bBaudRate<<8) | bCardType;
	  DEBUG(5,"dr->value(BaudRate, Cardtype)=%.4x\n",cardman->dr->value);
          cardman->dr->index=bStopBits;
          cardman->dr->length=0;
        #endif
	cardman->ctrl->dev=cardman->dev;
	xusb_submit_urb(cardman->ctrl);
	
	DEBUG(3,"<- set_card_parameter\n");
}

STATIC int set_protocol(usb_cardman_t *cardman, ptsreq_t *ptsreq){
 
  unsigned long tmp;

  DEBUG(3,"-> set_protocol\n");
  DEBUG(5,"ptsreq->protocol = %ld \nptsreq->Flags=0x%.2x\nptsreq->pts1=0x%.2x\nptsreq->pts2=0x%.2x\nptsreq->pts3=0x%.2x\n", 
          ptsreq->protocol, ptsreq->flags, ptsreq->pts1, ptsreq->pts2, ptsreq->pts3);

  // Fill PTS structure
  cardman->pts[0] = 0xff;
  cardman->pts[1] = 0x00;
  tmp = ptsreq->protocol;
  while((tmp=(tmp>>1))>0){cardman->pts[1]++;}
  cardman->proto=cardman->pts[1]; // Set new protocol
  cardman->pts[1]=(0x01<<4)|(cardman->pts[1]);
  
  DEBUG(5,"cardman->pts[1] = 0x%.2x\n", cardman->pts[1]);

  // Correct Fi/Di according to CM2020 Fi/Di table
  DEBUG(4,"Ta(1) from ATR is 0x%.2x, ptsreq->pts1=0x%.2x\n", cardman->ta1, ptsreq->pts1);
  if (cardman->pts[1] == 0x10){ // T=0
    DEBUG(4, "PTS for protocol T=0\n");
    cardman->pts[2] = 0x11;
  }else if (cardman->pts[1] == 0x11){ // T=1
    DEBUG(4, "PTS for protocol T=1\n");
    cardman->pts[2] = ucFIDITable_T1_FW1FW2[ptsreq->pts1&0x0F][(ptsreq->pts1>>4)&0x0F]; //set Fi/Di according to ATR TA(1)
  }else{
    // protocol unsupported
    DEBUG(3,"<- set_protocol: Protocol unsupported!\n");
    return -EINVAL;
  }
  // Calculate PCK character
  cardman->pts[3]=cardman->pts[0]^cardman->pts[1]^cardman->pts[2];
  
  
  DEBUG(4,"pts0 = %.2x\n", cardman->pts[0]);
  DEBUG(4,"pts1 = %.2x\n", cardman->pts[1]);
  DEBUG(4,"pts2 = %.2x\n", cardman->pts[2]);
  DEBUG(4,"pts3 = %.2x\n", cardman->pts[3]);
  
  if (test_bit(IS_INVERSE, (void *)&cardman->flags)){
    str_invert_revert(cardman->pts, 4);
  }

  // Set up URB
  
  cardman->cbuf[0]=cardman->pts[0];
  cardman->cbuf[1]=cardman->pts[1];
  cardman->cbuf[2]=cardman->pts[2];
  cardman->cbuf[3]=cardman->pts[3];
  DEBUG(5,"PTS=%.2x %.2x %.2x %.2x\n",
	cardman->cbuf[0],
	cardman->cbuf[1],
	cardman->cbuf[2],
	cardman->cbuf[3]);
  cardman->ctrl->transfer_buffer=cardman->cbuf;
  cardman->ctrl->transfer_buffer_length=4;
  cardman->op=CB_WRITE_PTS;
  cardman->ttl=cardman->ttl_hi=10; 
  cardman->ptsttl=100/*2000*/;
  cardman->irq->start_frame=-1;
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
    cardman->dr->bRequestType=CM2020_REQT_WRITE;
    cardman->dr->bRequest=CM2020_MODE_1;
    cardman->dr->wValue=0;
    cardman->dr->wIndex=0;
    cardman->dr->wLength=4;
  #else
    cardman->dr->requesttype=CM2020_REQT_WRITE;
    cardman->dr->request=CM2020_MODE_1;
    cardman->dr->value=0;
    cardman->dr->index=0;
    cardman->dr->length=4;
  #endif
  cardman->ctrl->dev=cardman->dev;
  DEBUG(5,"Submitting PTS request\n");
  xusb_submit_urb(cardman->ctrl);  
  DEBUG(3,"<- set_protocol\n");
  return 0;
}

STATIC void request_status(usb_cardman_t *cardman) {
	DEBUG(3,"-> request_status\n");
	submit_ctrl(cardman, CM2020_REQT_WRITE ,CM2020_GET_STATUS, 0, 0);
	DEBUG(3,"<- request_status\n");
}

STATIC void request_atr(usb_cardman_t *cardman, int  power_act) {
	DEBUG(3,"-> request_atr\n");
	submit_ctrl(cardman,CM2020_REQT_WRITE,CM2020_CARD_ON, (unsigned short)power_act, 0);
	DEBUG(3,"<- request_atr\n");
}

STATIC ssize_t cmu_read(struct file * filp,char *buf,size_t count,loff_t *ppos) {
register usb_cardman_t *cardman=(usb_cardman_t *)filp->private_data;
int retval = 0;
#ifdef CM2020_DEBUG
int i;
#endif
 DEBUG(2,"-> cmu_read\n");

	if ((cardman==NULL) || (cardman->dev==NULL))
		return -ENODEV;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 /* lock this object */
	 down(&cardman->sem);

 	/* verify that the device wasn't unplugged */
	 if (!cardman->present) {
		retval = -ENODEV;
		goto exit;
	 }
	#endif
	 
	if (count==0) {		/* according to manpage */
		retval = 0;
		goto exit;
	}
	
	count=min(count,cardman->rlen);
	copy_to_user(buf,cardman->rbuf,count);
#ifdef CM2020_DEBUG
	DEBUG(4,"cmu_read: ");
	for(i=0;i<cardman->rlen;i++) {
	  if (i%8==0) DEBUG(4,"\n");
          DEBUG(4,"%.2x ",cardman->rbuf[i]);
	}
	DEBUG(4,"\n");
#endif
	cardman->rlen=0;

	retval = count;

	memset(cardman->rbuf, 0xFF, count);

exit:
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 up(&cardman->sem);
	#endif
	 DEBUG(2,"<- read returns %i\n", retval);
	return retval;	
}

STATIC ssize_t cmu_write(struct file * filp,const char *buf,size_t count,loff_t *ppos) {
register usb_cardman_t *cardman=(usb_cardman_t *)filp->private_data;
#ifdef CM2020_DEBUG
int i;
#endif
ssize_t rc;
int retval = 0;

#ifdef CM2020_DEBUG
	DEBUG(4,"-> cmu_write\n");
#endif

        if ((cardman==NULL) || (cardman->dev==NULL)) {
		return -ENODEV;
	}

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 /* lock this object */
       	 down (&cardman->sem);

	/* verify that the device wasn't unplugged */
	 if (!cardman->present) {
		up (&cardman->sem);
		retval = -ENODEV;
		goto exit;
	 }
	#endif
		
	if (count==0) {		/* according to manpage */
		retval = 0;
		goto exit;
	}

	if (cardman->proto==0 && count<4){ /* T0 must have at least 4 bytes */
		DEBUG(4, "T0 has a min buffer of 4 bytes (count = %i)\n", count);
		retval = -EIO;
		goto exit;
	}

	/* T1 has a max of 48 bytes buffer */
	if (cardman->proto && count>261){
		DEBUG(4, "T1 command exceeds kernel modules buffer size (count = %i)\n", count);
		retval = -E2BIG;
		goto exit;
	}
	
	if (!test_bit(IS_ATR_VALID, (void *)&cardman->flags)){
		DEBUG(4, "write: card not ready\n");
		retval = -EIO;
		goto exit;
	}
	
	/* then, wait for the device and lock it */
	DEBUG(4,"write: wait for dev\n");
	while(test_and_set_bit(LOCK_DEVICE, (void *)&cardman->flags)) {
		DEBUG(4,"write: wait for dev: go to sleep\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up LOCK_DEVICE\n");
    wait_event_interruptible(cardman->waitq,(1 == 1));
//	wait_event_interruptible(cardman->waitq,!test_and_set_bit(LOCK_DEVICE, (void *)&cardman->flags));
#else
		interruptible_sleep_on(&cardman->waitq);
#endif
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			goto exit;
		}
	}
	DEBUG(4,"write: wait for dev: got it\n");

	/* we may now write to the device */
	cardman->rlen=0;
	rc=0;

	cardman->cbuf[4]=0;
	copy_from_user(cardman->cbuf,buf,count);

	#ifdef CM2020_DEBUG
	DEBUG(4,"cmu_write: ");
 	  for (i=0; i<count; i++){
	    if (i%8==0) DEBUG(4,"\n");
	    DEBUG(4,"%.2x ", cardman->cbuf[i]);
	  }
	  DEBUG(4,"\n");
	  DEBUG(4,"cardman->proto = 0x%.2x\n", cardman->proto);
	#endif //CM2020_DEBUG
	if (cardman->proto==0) {
		cardman->t0_data_len=cardman->cbuf[4];
		cardman->t0_data_off=5;
		cardman->t0_ins=cardman->cbuf[1];
		/* ReadT0 */
		if (count==5)
			/* Lc + SW1SW2 */
			cardman->t0_expected_reply_len=( ((unsigned char)cardman->cbuf[4]==0)?256:(unsigned char)cardman->cbuf[4] )+2;
		/* WriteT0 */
		else {
		  if (count != 4) {
		    /* check data consistency: Lc contains length
		     * of data bytes, thus 5(header) + Lc == total
		     * length
		     */
		    if ((count < 5+(unsigned char)cardman->cbuf[4]) || (5+(unsigned char)cardman->cbuf[4]+1 < count)) {
		      DEBUG(4,"-EBADR: buf[4]=%.2x nr=%d\n",
			    cardman->cbuf[4],count);
		      rc=-EBADR;
		      goto release_io;
		    }
		  }
		  cardman->t0_expected_reply_len=2;
		}
	}
	if (test_bit(IS_INVERSE, (void *)&cardman->flags)){
	  str_invert_revert(cardman->cbuf,count);
	}

	cardman->ctrl->transfer_buffer=cardman->cbuf;
	cardman->ttl=cardman->ttl_hi=TIMEOUT_LEN/cardman->bInterval;
	cardman->irq->start_frame=-1;
	if (cardman->proto==1) {
	  cardman->rlen=0;
	  cardman->t1_reply_len=0;
	  if (test_bit(T1MODE2, (void*)&cardman->flags) == 0){
	    DEBUG(4,"cardman->prot==T1\n");
	    cardman->bwt=TIMEOUT_LEN/cardman->bInterval;
	    cardman->ctrl->transfer_buffer_length=count;
	    cardman->op=CB_WRITE_T1;
            #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	      cardman->dr->bRequestType=CM2020_REQT_WRITE;
	      cardman->dr->bRequest=CM2020_MODE_1;
	      cardman->dr->wValue=0;
	      cardman->dr->wIndex=0;
	      cardman->dr->wLength=count;
	    DEBUG(4, "cardman->dr->wLength = %i\n", cardman->dr->wLength);
            #else
	      cardman->dr->requesttype=CM2020_REQT_WRITE;
	      cardman->dr->request=CM2020_MODE_1;
	      cardman->dr->value=0;
	      cardman->dr->index=0;
	      cardman->dr->length=count;
            #endif
	    cardman->ctrl->dev=cardman->dev;
	  }else{
	    DEBUG(4,"cardman->prot==T1 (transparent mode)\n");
	    // write T=1 package
	    cardman->bwt=TIMEOUT_LEN/cardman->bInterval;
	    cardman->ctrl->transfer_buffer_length=count;
	    cardman->op=CB_T1MODE2;
            #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	      cardman->dr->bRequestType=CM2020_REQT_WRITE;
	      cardman->dr->bRequest=CM2020_MODE_2;
	      cardman->dr->wValue=0;
	      cardman->dr->wIndex=0;
	      cardman->dr->wLength=count;
	    DEBUG(4, "cardman->dr->wLength = %i\n", cardman->dr->wLength);
            #else
	      cardman->dr->requesttype=CM2020_REQT_WRITE;
	      cardman->dr->request=CM2020_MODE_2;
	      cardman->dr->value=0;
	      cardman->dr->index=0;
	      cardman->dr->length=count;
            #endif
	    cardman->ctrl->dev=cardman->dev;
	  }
	}
	else if (cardman->proto==0) {
	  DEBUG(4,"cardman->proto=T0\n");
	  cardman->ctrl->transfer_buffer_length=5;
          #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	    cardman->dr->bRequestType=CM2020_REQT_WRITE;
          #else
	    cardman->dr->requesttype=CM2020_REQT_WRITE;
          #endif
	  if ((count>5) || (count == 4)) {
	    DEBUG(4, "cmu_write: WRITE_T0\n");
	    if (count == 4){
	      DEBUG(4, "cmu_write: fill up 4 byte APDU\n");
	      cardman->cbuf[4]=0;
	      count=5;
	    }
	    cardman->op=CB_WRITE_T0;
            #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	      cardman->dr->bRequest=CM2020_MODE_2;
	      cardman->dr->wValue=0;
	      cardman->dr->wIndex=0;
            #else
	      cardman->dr->request=CM2020_MODE_2;
	      cardman->dr->value=0;
	      cardman->dr->index=0;
            #endif
	  }
	  else {
	    DEBUG(4, "cmu_write: READ_T0\n");
	    cardman->op=CB_READ_T0;
            #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	      cardman->dr->bRequest=CM2020_MODE_3;
	      /* timeout */
	      cardman->dr->wValue=0x0a;	/* 2,56 sec */
	      // XXX FIX THIS: CWT from ATR.
	      // cardman->dr->value=0x08;	/* 2,048 sec */
	      /* INS | Lc */
	      cardman->dr->wIndex=(cardman->cbuf[1]<<8)|
		(cardman->t0_data_len);
	      DEBUG(4,"dr->index=%.4x\n",cardman->dr->wIndex);
            #else
	      cardman->dr->request=CM2020_MODE_3;
	      /* timeout */
	      cardman->dr->value=0x0a;	/* 2,56 sec */
	      // XXX FIX THIS: CWT from ATR.
	      // cardman->dr->value=0x08;	/* 2,048 sec */
	      /* INS | Lc */
	      cardman->dr->index=(cardman->cbuf[1]<<8)|
		(cardman->t0_data_len);
	      DEBUG(4,"dr->index=%.4x\n",cardman->dr->index);
            #endif
	      
	  }
          #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	    cardman->dr->wLength=5;
          #else
	    cardman->dr->length=5;
          #endif
	    cardman->ctrl->dev=cardman->dev;
	}
	set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	xusb_submit_urb(cardman->ctrl);
	
	cmwaitforOpFinish(cardman);

	/* wait for op to complete */
	if (cardman->proto==1){
	  if (cardman->bwt <= 0){
	    DEBUG(1, "BWT < 0 returning -ETIME\n");
	    rc=-ETIME;
	  }
	  else if (cardman->t1_reply_len==-1){
	    DEBUG(1, "t1_replylen == -1:  returning -EIO\n");
	    rc=-EIO;
	  }
	} else if ((cardman->proto==0) && (cardman->rlen==0)){
	  rc=-ETIME; 
	}
	if (test_bit(IS_INVERSE, (void *)&cardman->flags))
	  str_invert_revert(cardman->rbuf,cardman->rlen);
release_io:
	memset(cardman->cbuf, 0xFF, count);
	clear_bit(LOCK_DEVICE, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up LOCK_DEVICE\n");
	wake_up_interruptible(&cardman->waitq);
#else
	wake_up_interruptible(&cardman->waitq);
#endif

	DEBUG(4,"<- cmu_write (rc = %i, count = %i)\n", rc, count);

	/* return error or actually written bytes */
	retval = rc<0?rc:count;
exit:
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 up(&cardman->sem);
	#endif
	return retval;	
}

STATIC int cmu_ioctl(struct inode *inode,struct file *filp,
				unsigned int cmd,unsigned long arg) {
usb_cardman_t *cardman = (usb_cardman_t *)filp->private_data;
int size, rc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#else
 int i;
#endif
int retval = 0;

	DEBUG(2,"-> cmu_ioctl: minor=%d\n", MINOR(inode->i_rdev));
	
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 //ct!-- cardman = (usb_cardman_t *)filp->private_data;
	#else
	 i=MINOR(inode->i_rdev)-CM2020_MINOR;
	 if (i>=CM2020_MAX_DEV)
	 	return -ENODEV;
	 //ct!-- cardman=cardman_table[i];
	#endif

	if ((cardman==NULL) || (cardman->dev==NULL))
		return -ENODEV;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 /* lock this object */
	 down (&cardman->sem);

	 /* verify that the device wasn't unplugged */
	 if (!cardman->present) {
	 	retval = -ENODEV;
		goto exit;
	 }
	#endif
	
	if (_IOC_TYPE(cmd)!=CM_IOC_MAGIC) {
		retval = -EINVAL;
		goto exit;
	}
	if (_IOC_NR(cmd)>CM_IOC_MAXNR) {
		retval = -EINVAL;
		goto exit;
	}
	size =_IOC_SIZE(cmd);
	DEBUG(5,"ioctl:  IOC command size=%d\n",size);
	rc=0;
	if (_IOC_DIR(cmd)&_IOC_READ) {
		DEBUG(5,"ioctl: _ioc_dir=read\n");
		if (arg==0) rc=-EFAULT;
		else rc=verify_area(VERIFY_WRITE,(void *)arg,size);
	}
	if (_IOC_DIR(cmd)&_IOC_WRITE) {
		DEBUG(5,"ioctl: _ioc_dir=write\n");
		if (arg==0) rc=-EFAULT;
		else rc=verify_area(VERIFY_READ,(void *)arg,size);
	}
	if (rc) {
	  DEBUG(4, "<- ioctl (unknown comamnd)\n");
	  retval = rc;
	  goto exit;
	}

	rc = 0;

	DEBUG(4, "ioctl: Get device lock\n");
	while(test_and_set_bit(LOCK_DEVICE, (void *)&cardman->flags)) {
	  if (filp->f_flags==O_NONBLOCK) {
	  	retval = -ERESTARTSYS;
		goto exit;
	  }
	  DEBUG(4,"ioctl: wait for dev: go to sleep\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wait event LOCK_DEVICE\n");
	wait_event_interruptible(cardman->waitq,!test_and_set_bit(LOCK_DEVICE, (void *)&cardman->flags));
#else
	  interruptible_sleep_on(&cardman->waitq);
#endif
	  if (signal_pending(current)) {
	  	retval = -ERESTARTSYS;
		goto exit;
	  }
	}

	switch(cmd) {
	case CM_IOCGSTATUS:
	  {
	    unsigned char tmp;

	  DEBUG(4,"ioctl=CM_IOCGSTATUS\n");
 
	  set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	  cardman->op=CB_READ_STATUS;
	  cardman->ttl=cardman->ttl_hi;
	  request_status(cardman);
	  
	  cmwaitforOpFinish(cardman);
	  
	  tmp=0|(cardman->card_state>>6)
	    | (test_bit(IS_ATR_PRESENT, (void *)&cardman->flags)<<2)
	    | (test_bit(IS_ATR_VALID, (void *)&cardman->flags)<<3);
	  copy_to_user(arg, &(tmp), sizeof(unsigned char));
	  DEBUG(4,"card_state(pcsc coding) = %.2x\n",
		tmp);
	  DEBUG(4,"reader returns %.2x\n",cardman->card_state);
	  }
	  break;
	case CM_IOCGATR:
	  DEBUG(4,"ioctl=CM_IOCGATR\n");
	  {
	    atreq_t *atreq=(atreq_t *)arg;
	    ptsreq_t ptsreq;
	    int tmp, action;

	    copy_from_user(&action, &(atreq->power_act), sizeof(int) );
	    DEBUG(4, "ioctl: CM_IOCGATR: power_action = %i\n", action);

	    set_bit(DO_AUTOPPS, (void *)&cardman->flags);
	  start_pps:
	    DEBUG(4, "ioctl: set initial card parameter\n");
	    set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	    cardman->op=CB_SET_PARAMETER;
	    set_card_parameter(cardman,CM2020_CARD_ASYNC,CM2020_BAUDRATE_9600, 0x02);
	    
	    cmwaitforOpFinish(cardman);

	    DEBUG(4,"ioctl: request_status\n");
	    set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	    cardman->op=CB_READ_STATUS;
	    cardman->ttl=cardman->ttl_hi;
	    request_status(cardman);
	    
	    cmwaitforOpFinish(cardman);
	    
	    if ( (cardman->card_state != 0x00)){ // card inserted ? 
	      DEBUG(4, "ioctl: start request_atr\n");
	      set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	      cardman->op=CB_READ_ATR;
	      cardman->atr_len = 0;
	      cardman->ttl = cardman->ttl_hi = 120;
	      request_atr(cardman, action);

	      cmwaitforOpFinish(cardman);
	    }
	    
	    //Set card parameter to reflect inverse-reverse convention
	    if (test_bit(IS_ATR_VALID, (void *)&cardman->flags) && test_bit(IS_INVERSE, (void *)&cardman->flags)){
	      set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	      cardman->irq->start_frame=-1;
	      cardman->op=CB_SET_PARAMETER;
	      DEBUG(4,"ioctl: set_card_parameter to reflect inverse-reverse convention!\n");
	      set_card_parameter(cardman,CM2020_CARD_ASYNC,CM2020_BAUDRATE_9600 | CM2020_ODD_PARITY, 0x02);

	      cmwaitforOpFinish(cardman);

	    }
	    
	    // perform autoPPS
	    DEBUG(4, "*** Ta1 is 0x%.2x\n", cardman->ta1);
	    if (test_bit(IS_ATR_VALID, (void *)&cardman->flags) &&
                test_bit(DO_AUTOPPS, (void *)&cardman->flags) && 
                (cardman->ta1 != 0x11) &&
                !(test_bit(IS_ANY_T0, (void *)&cardman->flags) && test_bit(IS_ANY_T1, (void *)&cardman->flags))){
	      DEBUG(4,"ioctl: Performing AUTO PPS\n");
	      ptsreq.protocol = (0x01<<cardman->proto);
	      ptsreq.flags = 0x01;
	      ptsreq.pts1 = cardman->ta1;
	      ptsreq.pts2 = 0x00;
	      ptsreq.pts3 = 0x00;
	      set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	      set_protocol(cardman, &ptsreq);


	      cmwaitforOpFinish(cardman);

	      clear_bit(DO_AUTOPPS, (void *)&cardman->flags);
	      if (!test_bit(IS_ATR_VALID, (void *)&cardman->flags)){ 
		// something went wrong during auto PPS, so we power again using the default values
		DEBUG(4, "ioctl: Auto PPS failed: try again using  default values\n");
		goto start_pps;
	      }else{
		DEBUG(4, "ioctl: Auto PPS successful\n");
	      }	      
	    }//else: Card not powered, second try to power card failed or PSS not necessary
	    #ifdef CM2020_DEBUG
	    else{
	      DEBUG(4, "ioctl: No AUTOPPS:\n");
	      if (!test_bit(IS_ATR_VALID, (void *)&cardman->flags)){
		DEBUG(4, "    Card not powered, second try to power card failed or PSS not necessary\n");
	      }else if (!test_bit(DO_AUTOPPS, (void *)&cardman->flags)){
		DEBUG(4, "    Second try to power up card (previous AUTOPPS failed)\n");
	      }else if (cardman->ta1 == 0x11){
		DEBUG(4, "    PPS not necessary (ta1=0x11)\n");
	      }else if ( test_bit(IS_ANY_T0, (void *)&cardman->flags) && test_bit(IS_ANY_T1, (void *)&cardman->flags) ){
		DEBUG(4, "    Multiprotocol card (PCSCLite has to take care of PTS)\n");
	      }
	    } 
	    #endif //CM2020_DEBUG
	    
	    // reply
	    DEBUG(4, "ioctl: about to reply ATR flags = 0x%8x\n", cardman->flags);
	    if (!test_bit(IS_ATR_VALID, (void *)&cardman->flags)){
	      DEBUG(4,"ioctl: ATR invalid\n");
	      memset(&cardman->atr, 0, MAX_ATR);
	      cardman->atr_len=0;
	    }
	    #ifdef CM2020_DEBUG
	    else {
	      DEBUG(4,"ioctl: ATR valid\n");
	    }
	    #endif //CM2020_DEBUG
	    
	    // copy back ATR request data
	    tmp = cardman->atr_len;
	    copy_to_user(atreq->atr, cardman->atr, tmp);
	    copy_to_user(&(atreq->atr_len), &(tmp), sizeof(int));
	    copy_to_user(&(atreq->bIFSD), &(cardman->bIFSD), sizeof(unsigned char));
	    copy_to_user(&(atreq->bIFSC), &(cardman->bIFSC), sizeof(unsigned char));
	  }	    
	  break;
	case CM_IOCSPTS:
	  DEBUG(4,"ioctl=CM_IOCSPTS\n");	
	  {
	    ptsreq_t kernptsreq;
	    
	    copy_from_user(&kernptsreq, arg, sizeof(ptsreq_t));

	    if (kernptsreq.pts1==0x00)
		kernptsreq.pts1=cardman->ta1;
			   
	    if (test_bit(DO_AUTOPPS, (void*)&cardman->flags))
            { // PPS not already done automatically
	      DEBUG(4, "PTS: not already done, do it now\n");
	      DEBUG(4,"kernptsreq->pts1=0x%.2x\n", kernptsreq.pts1);
if ((cardman->ta1 == 0x11) && (cardman->proto == 0x00))
              {
	  DEBUG(4,"cardman->proto = 0x%.2x\n", cardman->proto);
	      DEBUG(4,"TA1 = 0x%x so do not PPS",cardman->ta1);
	      clear_bit(DO_AUTOPPS, (void*)&cardman->flags);
	  retval = 0;
	  goto release_io;
}
	      set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	      rc = set_protocol(cardman, &kernptsreq);
	      DEBUG(4,"Waiting for set_protocol to finish!");

	      cmwaitforOpFinish(cardman);
	      clear_bit(DO_AUTOPPS, (void*)&cardman->flags);
	    }else{
	      if (( (kernptsreq.protocol>>1) != cardman->proto) 
		  /*|| (kernptsreq.pts1 != cardman->ta1)*/){
		DEBUG(4, "PTS: values user wants to set are not equal to already applied settings\n");
		retval=-1;
	      }else{
		DEBUG(4, "PTS: settings already applied\n");
	      }
	    }
	  }
	  break;
	case CM_IOCSRDR:
	  DEBUG(4,"ioctl=CM_IOCSRDR (do nothing)\n");	
	  break;
	case CM_IOCARDOFF:
	  DEBUG(4,"ioctl=CM_IOCARDOFF\n");
	  
	  //turn off card
	  set_bit(IS_OP_PENDING, (void *)&cardman->flags);
	  cardman->op=CB_CARD_OFF;
	  submit_ctrl(cardman,CM2020_REQT_WRITE,CM2020_CARD_OFF, 0, 0);
	  DEBUG(4, "*** CB_CARD_OFF wait 50 msec\n");
	  mdelay(50);
	  DEBUG(4, "*** CB_CARD_OFF waiting 50 msec done\n");
	  cmwaitforOpFinish(cardman);

      	  break;
	case CM_IOSDBGLVL: /* set debug log level*/
	  {
	    int old_cm_debug=0;

	    old_cm_debug = cm_debug;
	    copy_from_user(&cm_debug, arg, sizeof(int));
	    if (old_cm_debug != cm_debug)
	      DEBUG(4, "Changed debug log level to %i\n", cm_debug);
	  }
	  break;
	default:
	  DEBUG(4,"<- ioctl returns EINVAL\n");
	  rc = -EINVAL;
	  break;
	}
release_io:	
	DEBUG(4, "ioctl: free device lock, schedule other requesters\n");
	clear_bit(LOCK_DEVICE, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    DEBUG(2,"wake up LOCK_DEVICE\n");
	wake_up_interruptible(&cardman->waitq);
#else
	wake_up_interruptible(&cardman->waitq);
#endif

	DEBUG(2, "<- ioctl return (%i)\n", rc);
	retval = rc;
exit:
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 up(&cardman->sem);
	#endif
	return retval;	
}
 
STATIC int cmu_open(struct inode *inode,struct file *filp) {
usb_cardman_t *cardman = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 int subminor;
 struct usb_interface *iface;
 struct usb_device *dev; //ct!--
#else
 int i;
#endif
int retval = 0;

 DEBUG(2,"-> cmu_open()\n");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	subminor = iminor(inode);
	down (&disconnect_sem);

	iface = usb_find_interface (&cmu_driver, subminor);
	if (!iface) {
		DEBUG(2,"<- iface == NULL: returning -ENODEV\n");
		retval = -ENODEV;
		goto exit;
	}
	cardman = usb_get_intfdata(iface);

	dev = interface_to_usbdev(iface);
	if (!dev) {
	  printk(KERN_WARNING MODULE_NAME "cmu_probe: did not find device for interface\n");
	  retval = -ENODEV;
	  goto exit;
	}
#else
	DEBUG(2,"-> cmu_open: major = %d,  minor=%d\n",
		MAJOR(inode->i_rdev), MINOR(inode->i_rdev)); 
	i=MINOR(inode->i_rdev)-CM2020_MINOR;
	if (i<0)
		printk(KERN_WARNING MODULE_NAME ": BUG! minor<CM2020_MINOR (%d<%d)\n",
			MINOR(inode->i_rdev),CM2020_MINOR);

	if ((i<0) || (i>=CM2020_MAX_DEV)) {
		retval = -ENODEV;
		goto exit;
	}
	cardman=cardman_table[i];
#endif
	
	if (cardman==NULL) {
	  DEBUG(2,"<- cardman == NULL: returning -ENODEV\n");
	  retval = -ENODEV;
	  goto exit;
	}
	if(cardman->dev==NULL){
	  DEBUG(2,"<- cardman->dev == NULL: retuning -ENODEV\n");
	  retval = -ENODEV;
	  goto exit;
	  
	}
	if (filp->f_flags & O_NONBLOCK){
	  DEBUG(2,"NONBLOCk flag set\n");
	  retval = -EAGAIN;
	  goto exit;
	}
	if (test_and_set_bit(IS_OPEN, (void *)&cardman->flags)==-1){
	  DEBUG(2,"device busy\n");
	  retval = -EBUSY;
	  goto exit;
	}
	DEBUG(4,"INC USE COUNT...");
        #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	  down(&cardman->sem);
	  ++cardman->open;
        #else
	   MOD_INC_USE_COUNT;
        #endif
	DEBUG(4,"done!\n");

	DEBUG(4,"Filling file struct\n");
	filp->private_data=cardman;
	DEBUG(4,"Setting atr_len\n");
	cardman->atr_len=0;
	DEBUG(4,"Setting owner\n");
	cardman->owner=current;

	//>>>> BEGIN FILL URBS
	/* WriteP0 */
      	DEBUG(4,"Setting up CTRL URB\n");        
	FILL_CONTROL_URB(cardman->ctrl,			/* urb           */
			 cardman->dev,			/* device        */
			 usb_sndctrlpipe(cardman->dev,0),/* pipe          */
			 (void *)cardman->dr,		/* setup request */
			 NULL, 0,			/* buffer,size   */
			 cardman_ctrl_callback,		/* completion    */
			 //NULL,
			 cardman				/* context       */
			 );
   
	/* ReadP0 */
	DEBUG(4,"Setting up RCTL URB\n");
	FILL_CONTROL_URB(cardman->rctl,			/* urb           */
			 cardman->dev,			/* device        */
			 usb_rcvctrlpipe(cardman->dev,0),/* pipe          */
			 (void *)cardman->dr,		/* setup request */
			 cardman->rcbuf, 0,
			 cardman_rctl_callback,		/* completion    */
			 cardman				/* context       */
			 );

	/* ReadP1 */
	DEBUG(4,"Setting up IRQ URB\n");
	FILL_INT_URB(cardman->irq,			/* urb           */
		     cardman->dev,			/* device        */
		     usb_rcvintpipe(cardman->dev,cardman->intendp),	/* pipe          */
		     cardman->ibuf,8,		/* buffer,size   */
		     cardman_irq_callback,		/* completion    */
		     cardman,			/* context       */
		     cardman->bInterval		/* intervall     */
		     );
	//<<<END FILL URBS


	DEBUG(4,"Submiting IRQ URB\n");
	cardman->irq->start_frame=-1;
	set_bit(IS_OP_PENDING, (void*)&cardman->flags);
	xusb_submit_urb(cardman->irq);
	
	clear_bit(LOCK_DEVICE, (void *)&cardman->flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	up (&cardman->sem);
#endif
	DEBUG(2,"<- cmu_open (OKAY)\n");

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 up (&disconnect_sem);
	 return retval;
	#else
	 return 0;
        #endif

exit:
	xusb_unlink_urb(cardman->irq);
	clear_bit(LOCK_DEVICE, (void *)&cardman->flags);
	clear_bit(IS_OPEN, (void *)&cardman->flags);
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 up (&cardman->sem);
	 --cardman->open;
	 up (&disconnect_sem);
	#else
	 MOD_DEC_USE_COUNT;
	#endif
	DEBUG(2,"<- cmu_open (FAILED)\n");
	return retval;
}

STATIC int cmu_close(struct inode *inode,struct file *filp) {
usb_cardman_t *cardman;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#else
 int i;
#endif

	DEBUG(2,"-> cmu_close\n");

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	#else
	 i=MINOR(inode->i_rdev)-CM2020_MINOR;
	 if (i>=CM2020_MAX_DEV){
	 	DEBUG(2,"<- cmu_close: wrong minor#, returning -ENODEV\n");
	 	return -ENODEV;
	 }
	#endif	 

	cardman=(usb_cardman_t *)filp->private_data; 
	if ((cardman==NULL) || (!test_bit(IS_OPEN, (void *)&cardman->flags))){
	  DEBUG(4," device not opened!\n");
	  DEBUG(2,"<- cmu_close: returning -ENODEV \n");
	  return -ENODEV;
	}
	
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 down(&cardman->sem);
	 if (cardman->open <=0 ) {
		DEBUG(4," device not opened (open<=0)!\n");
		DEBUG(2,"<- cmu_close: return -ENODEV \n");
		up(&cardman->sem);
		return -ENODEV;
	 }
        #endif
	
	DEBUG(4,"unlinking urbs\n");
	filp->private_data=NULL;
	cardman->op=CB_NOP;
	xusb_unlink_urb(cardman->irq);
	if (cardman->irq->status == -ENOENT) DEBUG(4,"IRQ URB killed!\n");
	xusb_unlink_urb(cardman->rctl);
	if (cardman->rctl->status == -ENOENT) DEBUG(4,"RCTL URB killed!\n");
	xusb_unlink_urb(cardman->ctrl);
	if (cardman->ctrl->status == -ENOENT) DEBUG(4,"CTRL URB killed!\n");
	cardman->flags=0;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 DEBUG(4,"Decrementing USE_COUNT\n");



//IsSleeping=1;
  //     wake_up_interruptible(&cardman->waitq);





	 --cardman->open;

 	 if (!cardman->present && !cardman->open) {
	 	/* the device was unplugged before the file was released */
	 	up (&cardman->sem);
	 	cardman_delete (cardman);
	 	return 0;
	 }

	 up (&cardman->sem);
	#else
	 if (cardman->dev == NULL){ // the device was unplugged before the file was released
	 	DEBUG(4,"device unplugged before file was released\n");
	 	cardman_delete(cardman);
	 }

	 DEBUG(4,"Decrementing USE_COUNT\n");
	 MOD_DEC_USE_COUNT;
#endif

	DEBUG(2,"<- cmu_close \n");
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC int cmu_probe(struct usb_interface *iface,
                        const struct usb_device_id *id) {
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
STATIC void *cmu_probe(struct usb_device *dev,unsigned int ifnum,
			const struct usb_device_id *id) {
  int i;
#else
STATIC void *cmu_probe(struct usb_device *dev,unsigned int ifnum) {
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 struct usb_device *dev = interface_to_usbdev(iface);
 int ifnum = 0; //ct!-- 
#endif	
struct usb_interface_descriptor *interface;
struct usb_endpoint_descriptor *endpoint0;
usb_cardman_t *cardman;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	if (!dev) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: did not find device for interface\n");
		return -ENODEV;
	}
	#endif
	DEBUG(2,"-> usb_cardman_probe idVendor=%x idProduct=%x ifnum=%d\n",
		dev->descriptor.idVendor,
		dev->descriptor.idProduct,
		ifnum);

	if (ifnum || 
	    ( !( (dev->descriptor.idVendor==0x076b) && (dev->descriptor.idProduct==0x0596 || dev->descriptor.idProduct==0x1784) ) && // OK 
	      !( (dev->descriptor.idVendor==0x08D4) && (dev->descriptor.idProduct==0x0009) )  ) || //FSC USB SmartCard Reader
	    ifnum) {
		printk(KERN_WARNING MODULE_NAME
			" cmu_probe: rejecting vend/prod/ifnum=%x/%x/%d\n",
			dev->descriptor.idVendor,dev->descriptor.idProduct,
			ifnum);
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENODEV;
		#else
		 return NULL;
		#endif
	}

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

	#else
	if (usb_set_configuration(dev,dev->config[0].bConfigurationValue)<0) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: usb_set_configuration failed\n");
		return NULL;
	}
	#endif

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 interface = &iface->altsetting[0].desc;
	 endpoint0 = &(iface->altsetting[0].endpoint[0].desc);
        #else
	 interface = dev->config[0].interface[ifnum].altsetting;
	 endpoint0=interface->endpoint;
	#endif
	
	if ((endpoint0->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) !=
		USB_ENDPOINT_XFER_INT) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: endpoint not of type XFER_INT\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENODEV;
		#else
		 return NULL;
		#endif
	}


	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	#else
	 for(i=0;i<CM2020_MAX_DEV;i++)
	 	if (cardman_table[i]==NULL)
	 		break;
	 if (i==CM2020_MAX_DEV) {
	 	printk(KERN_WARNING MODULE_NAME "cmu_probe: all devices in use\n");
	 	 return NULL;	/* all devices in use */
	 }
	 cardman_table[i]=NULL;
	#endif

	cardman=kmalloc(sizeof(struct usb_cardman),GFP_KERNEL);

 	if (cardman==NULL) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: can't malloc devices structure\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENOMEM;
		#else
		 return NULL;	/* all devices in use */
		#endif
	}

	memset(cardman,0,sizeof(struct usb_cardman));

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 init_MUTEX (&cardman->sem);
	 cardman->dev = dev;
	 cardman->interface = iface;
	#endif
	
	/* XXX MEMORY LEAK: free everything upon return! */

	cardman->ibuf=kmalloc(2*8+256+5,GFP_KERNEL);
	if (cardman->ibuf==NULL) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: can't malloc input buffer\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENOMEM;
		#else
		 return NULL;	/* all devices in use */
		#endif
	}
	cardman->cbuf=cardman->ibuf+8;
	cardman->rcbuf=cardman->cbuf+8;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	 cardman->dr=kmalloc(sizeof(struct usb_ctrlrequest),GFP_KERNEL);
        #else
	 cardman->dr=kmalloc(sizeof(devrequest),GFP_KERNEL);
	#endif
	if (cardman->dr==NULL) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: can't malloc device request\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENOMEM;
		#else
		 return NULL;	/* all devices in use */
		#endif
	}
	cardman->dev=dev;
	
	init_waitqueue_head(&cardman->waitq);

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	
	 cardman->ctrl=usb_alloc_urb(0, GFP_KERNEL);
	#else
	 cardman->ctrl=usb_alloc_urb(0);
	#endif
	if (cardman->ctrl==NULL) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: can't allocate control urb\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENOMEM;
		#else
		 return NULL;	/* all devices in use */
		#endif
	}
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	
	 cardman->rctl=usb_alloc_urb(0, GFP_KERNEL);
	#else
	 cardman->rctl=usb_alloc_urb(0);
	#endif
	if (cardman->rctl==NULL) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: can't allocate recv-control urb\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENOMEM;
		#else
		 return NULL;	/* all devices in use */
		#endif
	}
 
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	
	 cardman->irq=usb_alloc_urb(0, GFP_KERNEL);
	#else
	 cardman->irq=usb_alloc_urb(0);
	#endif
	if (cardman->irq==NULL) {
		printk(KERN_WARNING MODULE_NAME "cmu_probe: can't allocate interrupt urb\n");
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		 return -ENOMEM;
		#else
		 return NULL;	/* all devices in use */
		#endif
	}

	cardman->bInterval=endpoint0->bInterval;
	cardman->intendp=endpoint0->bEndpointAddress;

	printk(KERN_INFO MODULE_NAME " found a ");
	if (dev->descriptor.idProduct==0x0596) printk("OMNIKEY CardMan 2020\n");
	else if (dev->descriptor.idProduct==0x1784) printk("OMNIKEY CardMan 6020\n");
	else if (dev->descriptor.idProduct==0x0009) printk("FSC USB SmartCard Reader\n");
	else printk("-unknown reader\n"); // should never occur

        #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

 	 /* allow device read, write and ioctl */
	 cardman->present = 1;
	
	 /* we can register the device now, as it is ready */
	 usb_set_intfdata (iface, cardman);
	 if (usb_register_dev (iface, &cmu_class)) {
	 	/* something prevented us from registering this driver */
 		printk(KERN_WARNING MODULE_NAME "cmu_probe: Not able to get a minor for this device.\n");
 		usb_set_intfdata (iface, NULL);
		cardman_delete(cardman);
 		return -ENOMEM;
	 }
	 cardman->minor = iface->minor;
	#else
	 cardman_table[i]=cardman;
	#endif
	
	DEBUG(2, "<- probe\n");
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 return 0; 
	#else
	 return cardman;	/* all devices in use */
	#endif
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC void cmu_disconnect(struct usb_interface *iface) {
#else	
STATIC void cmu_disconnect(struct usb_device *dev,void *p) {
int i;
#endif
usb_cardman_t *cardman;
	DEBUG(2,"-> cmu_disconnect\n");  

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 down(&disconnect_sem);

	 cardman = usb_get_intfdata (iface);
	 usb_set_intfdata (iface, NULL);
	 
	 down(&cardman->sem);

	 /* give back our minor */
	 usb_deregister_dev (iface, &cmu_class);

	 //ct!-- terminate write
	 
	 cardman->present = 0;

	 up(&cardman->sem);
	#else
	for(i=0;i<CM2020_MAX_DEV;i++){
		if (cardman_table[i]==p) {
			cardman=cardman_table[i];
	#endif
			if(!test_bit(IS_OPEN, (void *)&cardman->flags)){
				DEBUG(4,"device  not opened clean up now\n");
				cardman_delete(cardman);
				#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
				
				#else
				 cardman_table[i]=NULL;
				#endif
			} else {
				DEBUG(4,"set usb_device == NULL signal unplugging\n");
				cardman->dev=NULL;
				cardman->op=CB_NOP;
			}
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	#else
			break;
		}
	}
	if (i>=CM2020_MAX_DEV)
		DEBUG(4,"device not found in cardman_table\n");
	#endif

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	 up (&disconnect_sem);
	#endif
	
	DEBUG(2,"<- cmu_disconnect\n");
}

static int __init cmu_init(void) {
	int result;
	
	printk(KERN_INFO "%s\n",version);
	result = usb_register(&cmu_driver);
	if (result) {
		printk(KERN_INFO "failed to register usb device (Error number %d)\n", result);
		return result;
	}
	return 0;
}

static void __exit cmu_exit(void) { 
	DEBUG(2,"cardman unloading\n");
	usb_deregister(&cmu_driver);
}

module_init(cmu_init);
module_exit(cmu_exit);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
 MODULE_AUTHOR("OMNIKEY AG");
 MODULE_DESCRIPTION("SmartCard driver for Omnikey CardMan USB 2020 and Dongle 6020");
 MODULE_LICENSE("Dual BSD/GPL");
#endif
